/*
 * ni-interface-filter.c
 * This file is part of NetworInador
 *
 * Copyright (C) 2021 - Félix Arreola Rodríguez
 *
 * NetworInador is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * NetworInador is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with NetworInador; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 */

#include "ni-interface-filter.h"

typedef struct _NIInterfaceFilterClass NIInterfaceFilterClass;
typedef struct _FilterRule FilterRule;

#define NI_INTERFACE_FILTER_CLASS(klass)     (G_TYPE_CHECK_CLASS_CAST ((klass), NI_TYPE_INTERFACE_FILTER, NIInterfaceFilterClass))
#define NI_IS_INTERFACE_FILTER_CLASS(klass)  (G_TYPE_CHECK_CLASS_TYPE ((klass), NI_TYPE_INTERFACE_FILTER))
#define NI_INTERFACE_FILTER_GET_CLASS(obj)   (G_TYPE_INSTANCE_GET_CLASS ((obj), NI_TYPE_INTERFACE_FILTER, NIInterfaceFilterClass))

typedef enum {
	FILTER_RULE_TYPE,
	FILTER_RULE_NAME,
	FILTER_RULE_INDEX,
	FILTER_RULE_CUSTOM
} FilterRuleType;

struct _NIInterfaceFilterClass {
	GInitiallyUnownedClass parent_class;
};

struct _NIInterfaceFilter {
	GInitiallyUnowned parent_instance;

	//NIInterfaceFilterType type;
	GSList *rules;
};

struct _FilterRule {
	FilterRuleType type;
	gboolean neg;
	
	union {
		guint index;
		guint type;
		gchar *name;
		struct {
			NIInterfaceFilterFunc func;
			gpointer data;
			GDestroyNotify notify;
		} custom;
	} u;
};

static void ni_interface_filter_finalize (GObject *object);

G_DEFINE_TYPE (NIInterfaceFilter, ni_interface_filter, G_TYPE_INITIALLY_UNOWNED)

static void ni_interface_filter_init (NIInterfaceFilter *object) {
	
}

static void ni_interface_filter_class_init (NIInterfaceFilterClass *class) {
	GObjectClass *gobject_class = G_OBJECT_CLASS (class);

	gobject_class->finalize = ni_interface_filter_finalize;
}

static void filter_rule_free (FilterRule *rule) {
	switch (rule->type) {
		case FILTER_RULE_NAME:
			g_free (rule->u.name);
			break;
		case FILTER_RULE_CUSTOM:
			if (rule->u.custom.notify) {
				rule->u.custom.notify (rule->u.custom.data);
			}
			break;
		//default:
			/* No importa */
	}
	
	g_slice_free (FilterRule, rule);
}

static void ni_interface_filter_finalize (GObject  *object) {
	NIInterfaceFilter *filter = NI_INTERFACE_FILTER (object);

	g_slist_free_full (filter->rules, (GDestroyNotify)filter_rule_free);

	//g_free (filter->name);

	G_OBJECT_CLASS (ni_interface_filter_parent_class)->finalize (object);
}

NIInterfaceFilter * ni_interface_filter_new (void) {
	NIInterfaceFilter *obj;
	obj = g_object_new (NI_TYPE_INTERFACE_FILTER, NULL);
	/*ni_interface_filter_set_type (obj, type);*/
	return obj;
}

/*void ni_interface_filter_set_type (NIInterfaceFilter *filter, const NIInterfaceFilterType type) {
	g_return_if_fail (NI_IS_INTERFACE_FILTER (filter));
	
	filter->type = type;
}

NIInterfaceFilterType ni_interface_filter_get_type (NIInterfaceFilter *filter) {
	g_return_if_fail (NI_IS_INTERFACE_FILTER (filter));
	
	return filter->type;
}*/

static void interface_filter_add_rule (NIInterfaceFilter *filter, FilterRule *rule) {
	filter->rules = g_slist_append (filter->rules, rule);
}

void ni_interface_filter_add_name (NIInterfaceFilter *filter, const gchar *name, gboolean negated) {
	FilterRule *rule;
	
	g_return_if_fail (NI_IS_INTERFACE_FILTER (filter));
	g_return_if_fail (name != NULL);
	
	rule = g_slice_new (FilterRule);
	rule->type = FILTER_RULE_NAME;
	rule->neg = negated;
	rule->u.name = g_strdup (name);
	
	interface_filter_add_rule (filter, rule);
}

void ni_interface_filter_add_index (NIInterfaceFilter *filter, const guint index, gboolean negated) {
	FilterRule *rule;
	
	g_return_if_fail (NI_IS_INTERFACE_FILTER (filter));
	
	rule = g_slice_new (FilterRule);
	rule->type = FILTER_RULE_INDEX;
	rule->neg = negated;
	rule->u.index = index;
	
	interface_filter_add_rule (filter, rule);
}

void ni_interface_filter_add_type (NIInterfaceFilter *filter, const guint type, gboolean negated) {
	FilterRule *rule;
	
	g_return_if_fail (NI_IS_INTERFACE_FILTER (filter));
	
	rule = g_slice_new (FilterRule);
	rule->type = FILTER_RULE_TYPE;
	rule->neg = negated;
	rule->u.type = type;
	
	interface_filter_add_rule (filter, rule);
}

void ni_interface_filter_add_custom (NIInterfaceFilter *filter, NIInterfaceFilterFunc func, gpointer data, GDestroyNotify notify) {
	FilterRule *rule;
	
	g_return_if_fail (NI_IS_INTERFACE_FILTER (filter));
	g_return_if_fail (func != NULL);
	
	rule = g_slice_new (FilterRule);
	rule->type = FILTER_RULE_CUSTOM;
	rule->neg = FALSE;
	rule->u.custom.func = func;
	rule->u.custom.data = data;
	rule->u.custom.notify = notify;
	
	interface_filter_add_rule (filter, rule);
}

gboolean ni_interface_filter_filter (NIInterfaceFilter *filter, const NIInterfaceFilterInfo *filter_info) {
	GSList *tmp_list;
	FilterRule *rule;
	
	/* Primero recorrer las reglas negadas */
	for (tmp_list = filter->rules; tmp_list; tmp_list = tmp_list->next) {
		rule = tmp_list->data;
		
		if (rule->neg == FALSE) continue;
		switch (rule->type) {
			case FILTER_RULE_INDEX:
				if (rule->u.index == filter_info->index) return FALSE;
				break;
			case FILTER_RULE_TYPE:
				if (rule->u.type == filter_info->type) return FALSE;
				break;
			case FILTER_RULE_NAME:
				if (g_regex_match_simple (rule->u.name, filter_info->name, 0, 0)) return FALSE;
				break;
		}
	}
	
	for (tmp_list = filter->rules; tmp_list; tmp_list = tmp_list->next) {
		rule = tmp_list->data;
		
		if (rule->neg == TRUE) continue;
		switch (rule->type) {
			case FILTER_RULE_INDEX:
				if (rule->u.index == filter_info->index) return TRUE;
				break;
			case FILTER_RULE_TYPE:
				if (rule->u.type == filter_info->type) return TRUE;
				break;
			case FILTER_RULE_NAME:
				if (g_regex_match_simple (rule->u.name, filter_info->name, 0, 0)) return TRUE;
				break;
			case FILTER_RULE_CUSTOM:
				if (rule->u.custom.func (filter_info, rule->u.custom.data)) return TRUE;
				else return FALSE;
				break;
		}
	}
	
	return FALSE;
}

