/*
 * link-types.h
 * This file is part of Network Inador
 *
 * Copyright (C) 2021 - Félix Arreola Rodríguez
 *
 * Network Inador is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Network Inador is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Network Inador; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 */

#ifndef __LINK_TYPES_H__
#define __LINK_TYPES_H__

typedef enum {
	/* Please don't interpret type numbers outside nm-platform and use functions
	 * like nm_platform_link_is_software() and nm_platform_supports_slaves().
	 *
	 * type & 0x10000 -> Software device type
	 * type & 0x20000 -> Type supports slaves
	 */

	/* No type, used as error value */
	NI_LINK_TYPE_NONE,

	/* Unknown type  */
	NI_LINK_TYPE_UNKNOWN,

	/* Hardware types */
	NI_LINK_TYPE_ETHERNET,
	NI_LINK_TYPE_INFINIBAND,
	NI_LINK_TYPE_OLPC_MESH,
	NI_LINK_TYPE_WIFI,
	NI_LINK_TYPE_WWAN_NET,   /* WWAN kernel netdevice */
	NI_LINK_TYPE_WIMAX,
	NI_LINK_TYPE_WPAN,
	NI_LINK_TYPE_6LOWPAN,

	/* Software types */
	NI_LINK_TYPE_BNEP = 0x10000,   /* Bluetooth Ethernet emulation */
	NI_LINK_TYPE_DUMMY,
	NI_LINK_TYPE_GRE,
	NI_LINK_TYPE_GRETAP,
	NI_LINK_TYPE_IFB,
	NI_LINK_TYPE_IP6TNL,
	NI_LINK_TYPE_IP6GRE,
	NI_LINK_TYPE_IP6GRETAP,
	NI_LINK_TYPE_IPIP,
	NI_LINK_TYPE_LOOPBACK,
	NI_LINK_TYPE_MACSEC,
	NI_LINK_TYPE_MACVLAN,
	NI_LINK_TYPE_MACVTAP,
	NI_LINK_TYPE_OPENVSWITCH,
	NI_LINK_TYPE_PPP,
	NI_LINK_TYPE_SIT,
	NI_LINK_TYPE_TUN,
	NI_LINK_TYPE_VETH,
	NI_LINK_TYPE_VLAN,
	NI_LINK_TYPE_VXLAN,
	NI_LINK_TYPE_WIREGUARD,

	/* Software types with slaves */
	NI_LINK_TYPE_BRIDGE = 0x10000 | 0x20000,
	NI_LINK_TYPE_BOND,
	NI_LINK_TYPE_TEAM,

	NI_LINK_TYPE_ANY = 0xffffffff,

} NILinkType;

typedef struct {
	const NILinkType link_type;
	const char *type_string;

	/* IFLA_INFO_KIND / rtnl_link_get_type() where applicable; the rtnl type
	 * should only be specified if the device type can be created without
	 * additional parameters, and if the device type can be determined from
	 * the rtnl_type.  eg, tun/tap should not be specified since both
	 * tun and tap devices use "tun", and InfiniBand should not be
	 * specified because a PKey is required at creation. Drivers set this
	 * value from their 'struct rtnl_link_ops' structure.
	 */
	const char *rtnl_type;
} LinkDesc;

static const LinkDesc linktypes[] = {
	{ NI_LINK_TYPE_NONE,          "none",        NULL },
	{ NI_LINK_TYPE_UNKNOWN,       "unknown",     NULL },

	{ NI_LINK_TYPE_ETHERNET,      "ethernet",    NULL },
	{ NI_LINK_TYPE_INFINIBAND,    "infiniband",  NULL },
	{ NI_LINK_TYPE_OLPC_MESH,     "olpc-mesh",   NULL },
	{ NI_LINK_TYPE_WIFI,          "wifi",        NULL },
	{ NI_LINK_TYPE_WWAN_NET,      "wwan",        NULL },
	{ NI_LINK_TYPE_WIMAX,         "wimax",       "wimax" },
	{ NI_LINK_TYPE_WPAN,          "wpan",        NULL },
	{ NI_LINK_TYPE_6LOWPAN,       "6lowpan",     NULL },

	{ NI_LINK_TYPE_BNEP,          "bluetooth",   NULL },
	{ NI_LINK_TYPE_DUMMY,         "dummy",       "dummy" },
	{ NI_LINK_TYPE_GRE,           "gre",         "gre" },
	{ NI_LINK_TYPE_GRETAP,        "gretap",      "gretap" },
	{ NI_LINK_TYPE_IFB,           "ifb",         "ifb" },
	{ NI_LINK_TYPE_IP6TNL,        "ip6tnl",      "ip6tnl" },
	{ NI_LINK_TYPE_IP6GRE,        "ip6gre",      "ip6gre" },
	{ NI_LINK_TYPE_IP6GRETAP,     "ip6gretap",   "ip6gretap" },
	{ NI_LINK_TYPE_IPIP,          "ipip",        "ipip" },
	{ NI_LINK_TYPE_LOOPBACK,      "loopback",    NULL },
	{ NI_LINK_TYPE_MACSEC,        "macsec",      "macsec" },
	{ NI_LINK_TYPE_MACVLAN,       "macvlan",     "macvlan" },
	{ NI_LINK_TYPE_MACVTAP,       "macvtap",     "macvtap" },
	{ NI_LINK_TYPE_OPENVSWITCH,   "openvswitch", "openvswitch" },
	{ NI_LINK_TYPE_PPP,           "ppp",         NULL },
	{ NI_LINK_TYPE_SIT,           "sit",         "sit" },
	{ NI_LINK_TYPE_TUN,           "tun",         "tun" },
	{ NI_LINK_TYPE_VETH,          "veth",        "veth" },
	{ NI_LINK_TYPE_VLAN,          "vlan",        "vlan" },
	{ NI_LINK_TYPE_VXLAN,         "vxlan",       "vxlan" },
	{ NI_LINK_TYPE_WIREGUARD,     "wireguard",   "wireguard" },

	{ NI_LINK_TYPE_BRIDGE,        "bridge",      "bridge" },
	{ NI_LINK_TYPE_BOND,          "bond",        "bond" },
	{ NI_LINK_TYPE_TEAM,          "team",        "team" },
};

#endif /* __LINK_TYPES_H__ */

