/*
 * main.c
 * This file is part of Network-inador
 *
 * Copyright (C) 2019, 2020 - Félix Arreola Rodríguez
 *
 * Network-inador is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Network-inador is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Network-inador; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 */

#include <stdio.h>
#include <stdlib.h>

#include <unistd.h>

#include <assert.h>

#include <asm/types.h>
#include <sys/socket.h>
#include <linux/netlink.h>

#include <glib.h>

#include <netlink/socket.h>
#include <netlink/msg.h>

#include <fcntl.h>

#include "network-inador.h"

/* Usados para salir en caso de una señal */
static int sigterm_pipe_fds[2] = { -1, -1 };

static void _sigterm_handler (int signum) {
	//fprintf (stderr, "SIGTERM SIGINT Handler\n");
	if (sigterm_pipe_fds[1] >= 0) {
		if (write (sigterm_pipe_fds[1], "", 1) == -1 ) {
			//fprintf (stderr, "Write to sigterm_pipe failed.\n");
		}
		close (sigterm_pipe_fds[1]);
		sigterm_pipe_fds[1] = -1;
	}
}

static gboolean _main_quit_handler (GIOChannel *source, GIOCondition cond, gpointer data) {
	GMainLoop *loop = (GMainLoop *) data;
	g_main_loop_quit (loop);
	
	return FALSE;
}

static void _main_setup_signal (void *loop) {
	struct sigaction act;
	sigset_t empty_mask;
	
	/* Preparar el pipe para la señal de cierre */
	if (pipe (sigterm_pipe_fds) != 0) {
		perror ("Failed to create SIGTERM pipe");
		sigterm_pipe_fds[0] = -1;
	}
	
	/* Instalar un manejador de señales para SIGTERM */
	sigemptyset (&empty_mask);
	act.sa_mask    = empty_mask;
	act.sa_flags   = 0;
	act.sa_handler = &_sigterm_handler;
	if (sigaction (SIGTERM, &act, NULL) < 0) {
		perror ("Failed to register SIGTERM handler");
	}
	
	if (sigaction (SIGINT, &act, NULL) < 0) {
		perror ("Failed to register SIGINT handler");
	}
	
	if (sigterm_pipe_fds[0] != -1) {
		fcntl (sigterm_pipe_fds[0], F_SETFD, 1);
		fcntl (sigterm_pipe_fds[1], F_SETFD, 1);
		GIOChannel *io;
		
		io = g_io_channel_unix_new (sigterm_pipe_fds[0]);
		g_io_channel_set_close_on_unref (io, TRUE);
		
		g_io_add_watch (io, G_IO_IN | G_IO_PRI | G_IO_HUP | G_IO_ERR, _main_quit_handler, loop);
		g_io_channel_unref (io);
	}
}

typedef struct _NetworkInadorGLibIOClosure {
	NetworkInadorInputFunc function;
	guint result;
	gpointer data;
} NetworkInadorGLibIOClosure;

static void glib_io_destroy (gpointer data) {
	g_free (data);
}

gboolean glib_io_invoke (GIOChannel *source, GIOCondition condition, gpointer data) {
	NetworkInadorGLibIOClosure *closure = data;
	int cond = 0;

	if (condition & G_IO_IN) {
		cond |= POLLIN;
	}
	if (condition & G_IO_OUT) {
		cond |= POLLOUT;
	}

	closure->function (closure->data, g_io_channel_unix_get_fd (source), cond);
	
	return TRUE;
}

unsigned int glib_input_add (int fd, int cond, NetworkInadorInputFunc func, void *user_data) {
	GIOChannel *channel;
	NetworkInadorGLibIOClosure *closure = g_new0 (NetworkInadorGLibIOClosure, 1);
	GIOCondition gcond = 0;
	
	closure->function = func;
	closure->data = user_data;
	
	if (cond == POLLIN) {
		gcond = G_IO_IN | G_IO_PRI | G_IO_HUP | G_IO_ERR;
	} else if (cond == POLLOUT) {
		gcond = G_IO_OUT | G_IO_PRI | G_IO_HUP | G_IO_ERR;
	}
	
	channel = g_io_channel_unix_new (fd);
	closure->result = g_io_add_watch_full (channel, G_PRIORITY_DEFAULT, cond, glib_io_invoke, closure, glib_io_destroy);
	
	g_io_channel_unref(channel);
	return closure->result;
}

struct NetworkInadorOps network_ops = {
	.input_add = glib_input_add,
	.input_remove = g_source_remove,
};

int main (int argc, char *argv[]) {
	NetworkInadorHandle *handle;
	GMainLoop *loop = NULL;

#if !defined(GLIB_VERSION_2_36)
    g_type_init ();
#endif
	
	handle = network_inador_init_handle (&network_ops);
	
	/* Dejar que la librería maneje los procesos hijos muertos */
	network_inador_setup_child_handler (handle);
	
	loop = g_main_loop_new (NULL, FALSE);
	
	_main_setup_signal (loop);
	
	interfaces_dhcp_client_run (handle, 3, 1, 1);
	/*IPAddr addr;
	inet_pton (AF_INET, "192.19.5.6", &addr.sin_addr);
	addr.family = AF_INET;
	addr.prefix = 25;
	addr.flags = 128;
	addr.scope = 0;
	
	ip_address_add_ip (&handle, 3, &addr);*/
	
	/*GList *g;
	Interface *x;
	for (g = handle.interfaces; g != NULL; g = g->next) {
		x = (Interface *) g->data;
		
		if (x->index == 2 && x->address != NULL) {
			ip_address_del_ip (&handle, x->index, x->address->data);
		}
	}*/
	
	//interfaces_bridge_create (&handle, "br0");
	//interfaces_bridge_set_master_interface (&handle, 4, 5);
	//interfaces_bridge_remove_master_interface (&handle, 5);
	//interfaces_bridge_delete (&handle, 4);
	
	//interfaces_dhcp_client_run (&handle, 2, IFACE_BUSYBOX_UDHCPC, DHCP_CLIENT_FLAG_AUTO_RESTART);
	
	#if 0
	// Prueba de agregar ruta
	Route route;
	GList nhl1, nhl2;
	RouteNH nh1, nh2;
	memset (&route, 0, sizeof (route));
	route.family = AF_INET;
	route.type = RTN_UNICAST;
	route.table = 0;
	inet_pton (AF_INET, "10.8.9.0", &route.dest);
	route.prefix = 24;
	
	route.protocol = RTPROT_STATIC;
	route.tos = 0;
	route.priority = 400;
	route.nexthops = &nhl1;
	
	memset (&nh1, 0, sizeof (nh1));
	memset (&nh2, 0, sizeof (nh2));
	nhl1.prev = NULL;
	nhl1.next = &nhl2;
	nhl2.prev = &nhl1;
	nhl2.next = NULL;
	nhl1.data = &nh1;
	nhl2.data = &nh2;
	
	inet_pton (AF_INET, "172.22.201.1", &nh1.gw);
	inet_pton (AF_INET, "172.22.201.2", &nh2.gw);
	nh1.out_index = 0;
	nh2.out_index = 0;
	nh1.nh_weight = 4;
	nh2.nh_weight = 9;
	
	routes_add (&handle, &route);
	#endif
	
	#if 0
	// Prueba de eliminar una ruta
	Route route;
	memset (&route, 0, sizeof (route));
	route.family = AF_INET;
	route.table = 0;
	inet_pton (AF_INET, "172.18.5.0", &route.dest);
	route.prefix = 24;
	
	route.tos = 0x18;
	
	routes_del (&handle, &route);
	
	interfaces_veth_create (&handle, "abc", "def");
	#endif
	/*interfaces_change_set_down (&handle, 4);
	int ret = interfaces_set_sysctl_ipv6_params (&handle, 4, DEVCONF_DISABLE_IPV6, (uint32_t) 1, -1);
	printf ("Ret sel sysctl: %i\n", ret);
	interfaces_change_set_up (&handle, 4);*/
	
	g_main_loop_run (loop);
	
	network_inador_destroy_handle (handle);
	
	g_main_loop_unref (loop);
	
	return 0;
}
