/*
 * resolv_conf_helper.c
 * This file is part of NetworkInador
 *
 * Copyright (C) 2023 - Félix Arreola Rodríguez
 *
 * NetworkInador is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * NetworkInador is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with NetworkInador; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <getopt.h>

#include <unistd.h>
#include <fcntl.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <arpa/inet.h>

#include "flist.h"

#include "network-inador-manager.h"
#include "resolv_conf_defs.h"
#include "struct_addr_union.h"
#include "resolv_conf_parser.h"
#include "utils.h"

#define COMMAND_SOCKET_PATH "/tmp/network-inador.socket"

static const struct option long_options[] = {
	{"enable-updates", 0, NULL, 'z' },
	{"disable-updates", 0, NULL, 'y' },
	{"updates-are-enabled", 0, NULL, 'x' },
	{"create-runtime-directories", 0, NULL, 'w' },
	{"wipe-runtime-directories", 0, NULL, 'v' },
	{NULL, 0, NULL, 0}
};

const char *short_options = "a:d:u";

int wait_for_ack_or_error (int s) {
	unsigned char buffer[128];
	int ret;
	
	ret = recv (s, buffer, sizeof (buffer), 0);
	
	if (ret < 3) {
		return -1;
	}
	
	if (buffer[0] == NET_INADOR_TYPE_RESPONSE_ERROR) {
		return -1;
	}
	
	if (buffer[0] == NET_INADOR_TYPE_RESPONSE && buffer[1] == NET_INADOR_RESPONSE_EXECUTED) {
		return 0;
	}
	
	return -1;
}

FList * helper_parser_parse_input (FList *all_entries, FILE *fd, int origin) {
	char buffer[8192], value[2048];
	ResolvConfEntry *entry = NULL;
	FList *pos, *next;
	int g, len;
	struct_addr direccion;
	int type;
	int family;
	int order;
	
	order = 0;
	while (fgets (buffer, sizeof (buffer), fd), feof (fd) == 0) {
		utils_trim (buffer);
		utils_trim_comment (buffer);
		if (buffer[0] == 0) continue;
		
		type = resolv_conf_parse_line (buffer, &family, &direccion, value, sizeof (value));
		
		if (type == -1) {
			/* Linea no válida de acuerdo al parser, ignorar */
			continue;
		}
		
		/* Desde el resolvconf helper, siempre anexar al final de la lista */
		entry = (ResolvConfEntry *) malloc (sizeof (ResolvConfEntry));
		
		entry->resolv_type = type;
		entry->origin = origin;
		
		entry->ns_family = family;
		memcpy (&entry->nameserver, &direccion, sizeof (direccion));
		strncpy (entry->value, value, sizeof (entry->value));
		
		entry->owner_interface_index = 0;
		entry->owner_prog[0] = 0;
		
		all_entries = f_list_append (all_entries, entry);
		entry->tagged = 1;
		entry->file_order = order++;
	}
	
	return all_entries;
}

void send_entries (int s, FList *entries, const char *iface_prog) {
	char buffer[2048];
	int pos, len, pos_name_count;
	FList *g;
	int n;
	ResolvConfEntry *entry;
	
	buffer[0] = NET_INADOR_TYPE_COMMAND;
	buffer[1] = NET_INADOR_COMMAND_RESOLVCONF_FEED;
	
	/* El segundo byte es el nombre seguido del nombre en sí */
	len = strlen (iface_prog);
	buffer[2] = len;
	
	memcpy (&buffer[3], iface_prog, len);
	
	pos_name_count = 3 + len;
	
	/* Contar los nameservers que tengo */
	g = entries;
	n = 0;
	pos = pos_name_count + 1;
	while (g != NULL) {
		entry = (ResolvConfEntry *) g->data;
		
		if (entry->resolv_type == RESOLV_TYPE_NAMESERVER) {
			n++;
			buffer[pos] = entry->ns_family;
			pos++;
			if (entry->ns_family == AF_INET) {
				memcpy (&buffer[pos], &entry->nameserver, 4);
				pos += 4;
			} else if (entry->ns_family == AF_INET6) {
				memcpy (&buffer[pos], &entry->nameserver, 16);
				pos += 16;
			}
		}
		
		g = g->next;
	}
	
	buffer[pos_name_count] = n;
	
	send (s, buffer, pos, 0);
	
	if (wait_for_ack_or_error (s) < 0) {
		return;
	}
}

void send_remove_entries (int s, const char *iface_prog) {
	char buffer[2048];
	int pos, len;
	
	buffer[0] = NET_INADOR_TYPE_COMMAND;
	buffer[1] = NET_INADOR_COMMAND_RESOLVCONF_REMOVE;
	
	/* El segundo byte es el nombre seguido del nombre en sí */
	len = strlen (iface_prog);
	buffer[2] = len;
	
	memcpy (&buffer[3], iface_prog, len);
	
	send (s, buffer, 3 + len, 0);
	
	if (wait_for_ack_or_error (s) < 0) {
		return;
	}
}

int main (int argc, char *argv[]) {
	int next_option;
	const char *iface_prog = NULL;
	int action = 0;
	FList *entries;
	int s, ret;
	struct sockaddr_un path_dest;
	
	do {
		next_option = getopt_long (argc, argv, short_options, long_options, NULL);
		
		switch (next_option) {
			case 'z':
			case 'y':
			case 'x':
			case 'w':
			case 'v':
				/* Ignorar */
				break;
			case 'a':
				action = 'a';
				iface_prog = optarg;
				break;
			case 'd':
				action = 'd';
				iface_prog = optarg;
				break;
			case 'u':
				break;
			case '?':
				/* Opción incorrecta */
				break;
			case -1:
				/* Fin de opciones */
				break;
		}
	} while (next_option != -1);
	
	if (action == 0) {
		return 0;
	}
	
	/* Validar el nombre del argumento del -a o del -d */
	if (iface_prog == NULL || iface_prog[0] == 0) {
		fprintf (stderr, "interface record name empty\n");
		return -1;
	}
	
	if (strchr (iface_prog, '/') != NULL) {
		/* Diagonal no permitida */
		fprintf (stderr, "Slash not allowed in interface record name\n");
		return -1;
	}
	
	if (strchr (iface_prog, ' ') != NULL) {
		fprintf (stderr, "Space not allowed in interface record name\n");
		return -1;
	}
	
	if (iface_prog[0] == '.') {
		fprintf (stderr, "Space not allowed in interface record name\n");
		return -1;
	}
	
	if (iface_prog[0] == '-') {
		fprintf (stderr, "Initial hyphen not allowed in interface record name\n");
		return -1;
	}
	
	if (iface_prog[0] == '~') {
		fprintf (stderr, "Initial tilde not allowed in interface record name\n");
		return -1;
	}
	
	if (strlen (iface_prog) > 255) {
		fprintf (stderr, "Interface too long\n");
		return -1;
	}
	
	/* Intentar abrir el socket al network inador */
	s = socket (AF_UNIX, SOCK_SEQPACKET | SOCK_CLOEXEC, 0);

	if (s < 0) {
		return 1;
	}
	
	path_dest.sun_family = AF_UNIX;
	strncpy (path_dest.sun_path, COMMAND_SOCKET_PATH, sizeof (path_dest.sun_path));
	ret = connect (s, (struct sockaddr *) &path_dest, sizeof (path_dest));
	
	if (ret < 0) {
		perror ("Connect");
		close (s);
		return 1;
	}
	
	if (action == 'a') {
		/* Parsear el archivo primero */
		entries = helper_parser_parse_input (NULL, stdin, RESOLV_ORIGIN_RESOLVCONF);
		
		/* Preparar un paquete de datos para enviar via el socket */
		send_entries (s, entries, iface_prog);
	} else if (action == 'd') {
		send_remove_entries (s, iface_prog);
	}
	
	close (s);
	
	return 0;
}

